// Define Query utility functions
Query.trend = function(view,select,where,interval,group) {
  var query = new Query();
  query.view     = view;
  query.select   = select;
  if(where != null) query.where = where;
  query.interval = interval;
  query.group    = group;
  return query;
}
Query.topN = function(view,select,where,interval,sort,n) {
  var query = new Query();
  query.view     = view;
  query.select   = select;
  if(where != null) query.where = where;
  query.interval = interval;
  if(sort != null) query.sort     = sort;
  if(n != null) query.truncate = n;
  return query;
}

Query.timeRange = function(interval,tz,view) {
  var query = new Query();
  query.view = view ? view : 'historytrmx';
  query.interval = interval;
  if(tz) query.tz = tz;
  return query.timeRange();
}

// Define class for storing query results
function Table(cnames,ctypes,cdata) {
  if(cnames && ctypes) {
    this.cnames = cnames;
    this.ctypes = ctypes;
    if(cdata) {
      this.cdata = cdata;
      this.ncols = cdata.length;
      this.nrows = cdata[0].length;
    }
    else {
      this.cdata = new Array();
      this.ncols = cnames.length;
      this.nrows = 0;
      for(var c = 0; c < this.ncols; c++) {
        this.cdata[c] = new Array();
      }
    }
  }
  else {
    this.cnames = new Array();
    this.ctypes = new Array();
    this.cdata = new Array();
    this.nrows = 0;
    this.ncols = 0;
  }
}

Table.create = function(cnames,ctypes,cdata) {
  return new Table(cnames,ctypes,cdata);
}

Table.prototype.addColumn = function(cname,ctype,data) {
  this.cnames.push(cname);
  this.ctypes.push(ctype);
  this.cdata.push(data);
  this.ncols++;
  if(data.length > this.nrows) this.nrows = data.length;
}

Table.prototype.cell = function(r,c) {
  return this.cdata[c][r];
}

Table.prototype.row = function(r) {
  var row = new Array();
  for(var c = 0; c < this.ncols; c++) {
    row[c] = this.cdata[c][r];
  }
  return row;
}

Table.prototype.column = function(c) {
  return this.cdata[c];
}

Table.prototype.insertColumn = function(cname,ctype,data,c) {
  this.cnames.splice(c,0,cname);
  this.ctypes.splice(c,0,ctype);
  this.cdata.splice(c,0,data);
  this.ncols++;
  if(data.length > this.nrows) this.nrows = data.length;
}

Table.prototype.deleteColumn = function(c) {
  this.cnames.splice(c,1);
  this.ctypes.splice(c,1);
  this.cdata.splice(c,1);
  this.ncols--;
  if(this.ncols == 0) this.nrows = 0;
}

Table.prototype.scaleColumn = function(c, factor) {
  for(var r = 0; r < this.nrows; r++) {
    this.cdata[c][r] *= factor;
  }
}

Table.prototype.addRow = function(row) {
  for(var c = 0; c < this.ncols; c++) {
    this.cdata[c].push(row[c]);
  }
  this.nrows++;
}

Table.prototype.deleteRow = function(r) {
  for(var c = 0; c < this.ncols; c++) {
    this.cdata[c].splice(r,1);
  }
  this.nrows--;
}

Table.prototype.truncateRows = function(limit) {
  if(limit < this.nrows && limit >= 0) {
    for(var c = 0; c < this.cdata.length; c++) this.cdata[c].length = limit;
    this.nrows = limit;
  }
}

Table.prototype.sort = function(sortcol,reverse) {
  var sortarray = new Array();
  for(var r = 0; r < this.nrows; r++) {
     var el = new Array(this.cdata[sortcol][r],r);
     sortarray.push(el);
  }
  if(reverse) sortarray.sort(function(a,b) {if(b[0] > a[0]) return 1; if(b[0] < a[0]) return -1; if(a[0] == b[0]) return 0; if(a[0] && !b[0]) return 1; if(b[0] && !a[0]) return -1; return 0;});
  else sortarray.sort(function(a,b) {if(a[0] > b[0]) return 1; if(a[0] < b[0]) return -1; if(a[0] == b[0]) return 0; if(!a[0] && b[0]) return 1; if(!b[0] && a[0]) return -1; return 0;});

  for(c = 0; c < this.ncols; c++) {
    var newcol = new Array();
    for(r = 0; r < this.nrows; r++) {
      var oldrow = sortarray[r][1];
      var val = this.cdata[c][oldrow];
      newcol[r] = val;
    }
    this.cdata[c] = newcol;
  }
}

Table.prototype.pivotTime = function(time,sub,value,otherflag) {
  var includeOther = otherflag;
  // first build the set of unique sub category values
  var hash = new Object();
  for(var r = 0; r < this.nrows; r++) {
    var c = this.cdata[sub][r];
    if(c != null || includeOther) {
      var sum = hash[c];
      if(sum == null) sum = 0;
      hash[c] = sum + this.cdata[value][r];
    }
  }

  // suppress other if zero
  if(includeOther && !hash['null']) {
    var ngroups = 0;
    for(var key in hash) ngroups++;
    // need to keep at least one group so that the time series exists
    if(ngroups > 1) {
      delete hash['null'];
      includeOther = false;
    }
  }

  var subcols = new Array();
  for(var key in hash) {
    subcols.push(key);
  }
  subcols.sort(
    function(a,b) {
      if(a == 'null' && b == 'null') return 0;
      if(a == 'null') return 1;
      if(b == 'null') return -1;
      var vala = hash[a];
      var valb = hash[b];
      return valb - vala;
    }
  );
  // build another hash that will map subname to subcol id
  // and initialize new column data arrays
  var new_cnames = new Array();
  var new_ctypes = new Array();
  var new_cdata = new Array();
  var new_pvttypes = new Array();
  new_cnames.push(this.cnames[time]);
  new_ctypes.push(this.ctypes[time]);
  new_pvttypes.push("string");
  new_cdata.push(new Array()); // time column
  var subidx = new Array();
  for(var i = 0; i < subcols.length; i++) {
    var idx = i + 1; // time column is idx = 0
    subidx[subcols[i]] = idx;
    new_cnames.push(subcols[i]);
    new_ctypes.push(this.ctypes[value]);
    new_pvttypes.push(this.ctypes[sub]);
    new_cdata.push(new Array());
  }

  // finally ready to traverse the table and fill in the
  // values

  var new_t;
  var new_r;
  for(var r = 0; r < this.nrows; r++) {
    var t = this.cdata[time][r];
    if(r == 0) {
       new_t = t;
       new_r = 0;
       new_cdata[0][new_r] = new_t;
       for(var c = 1; c <= subcols.length; c++) {
         new_cdata[c][new_r] = 0;
       }
    }
    else if(new_t - t != 0) {
      new_t = t;
      new_r++;
      new_cdata[0][new_r] = new_t;
      for(var c = 1; c <= subcols.length; c++) {
        new_cdata[c][new_r] = 0;
      }
    }
    var val = this.cdata[value][r];
    var subkey = this.cdata[sub][r];
    if(subkey != null || includeOther) new_cdata[subidx[subkey]][new_r] = val;
  }
  if(includeOther && (new_cnames[new_cnames.length - 1] == 'null')) {
     new_cnames[new_cnames.length - 1] = null;
     new_pvttypes[new_ctypes.length - 1] = 'other';
  }
  var newtable = new Table();
  newtable.cnames = new_cnames;
  newtable.ctypes = new_ctypes;
  newtable.pvttypes = new_pvttypes;
  newtable.cdata = new_cdata;
  newtable.ncols = newtable.cdata.length;
  newtable.nrows = newtable.cdata[0].length;
  newtable.start = this.start;
  newtable.end   = this.end;
  newtable.group = this.group;
  return newtable;
}

Table.prototype.pivotCategory = function(cat,sub,value) {
  // first build the set of unique sub category values
  var hash = new Array();
  for(var r = 0; r < this.nrows; r++) {
    var c = this.cdata[sub][r];
    if(c != null) {
    var sum = hash[c];
    if(sum == null) sum = 0;
    hash[c] = sum + this.cdata[value][r];}
  }
  var subcols = new Array();
  for(var key in hash) {
    subcols.push(key);
  }
  subcols.sort(
    function(a,b) {
      var vala = hash[a];
      var valb = hash[b];
      return valb - vala;
    }
  );
  // build another hash that will map subname to subcol id
  // and initialize new column data arrays
  var new_cnames = new Array();
  var new_ctypes = new Array();
  var new_pvttypes = new Array();
  var new_cdata = new Array();
  new_cnames.push(this.cnames[cat]);
  new_ctypes.push(this.ctypes[cat]);
  new_pvttypes.push("string");
  new_cdata.push(new Array()); // category column
  var subidx = new Array();
  for(var i = 0; i < subcols.length; i++) {
    var idx = i + 1; // cat column is idx = 0
    subidx[subcols[i]] = idx;
    new_cnames.push(subcols[i]);
    new_ctypes.push(this.ctypes[value]); 
    new_pvttypes.push(this.ctypes[sub]);
    new_cdata.push(new Array());
  }

  // finally ready to traverse the table and fill in the
  // values
  var rowhash = new Array();
  var new_r = 0;
  for(var r = 0; r < this.nrows; r++) {
    var c = this.cdata[cat][r];
    var row = rowhash[c];
    if(row == null) {
       row = new_r;
       new_cdata[0][row] = c;
       for(var s = 1; s <= subcols.length; s++) {
         new_cdata[s][new_r] = 0;
       }
       new_r++;
    }
    var val = this.cdata[value][r];
    var subkey = this.cdata[sub][r];
    if(subkey != null) new_cdata[subidx[subkey]][row] = val;
  }
  var newtable = new Table();
  newtable.cnames = new_cnames;
  newtable.ctypes = new_ctypes;
  newtable.pvttypes = new_pvttypes;
  newtable.cdata = new_cdata;
  newtable.ncols = newtable.cdata.length;
  newtable.nrows = newtable.cdata[0].length;
  newtable.start = this.start;
  newtable.end = this.end;
  newtable.end = this.group;
  return newtable;
}

Table.prototype.previousRank = function(prev,nkeys) { 
  var rank = new Array();
  if(this.ncols == prev.ncols) { 
    var prevRank = new Array();
    for(var r = 0; r < prev.nrows; r++) {
      var keystr = "";
      for(var c = 0; c < nkeys; c++) {
        if(c != 0) keystr += ",";
        if(prev.cell(r,c)) keystr += prev.cell(r,c);
      }
      if(keystr.length >= nkeys) prevRank[keystr] = r + 1;
    }
    for(var r = 0; r < this.nrows; r++) {
      var keystr = "";
      for(var c = 0; c < nkeys; c++) {
        if(c != 0) keystr += ",";
        keystr += this.cell(r,c);
      }
      if(prevRank[keystr]) rank[r] = prevRank[keystr];
      else rank[r] = null;
    }
  }
  return rank;
}

Table.prototype.addressesToNames = function(allFlag,dnsTimeout) {
  var n = new Network();
  for(var c = 0; c < this.ncols; c++) {
    if("address" == this.ctypes[c]) {
      var col = this.cdata[c];
      var map = n.dnsMap(col,allFlag,dnsTimeout);
      for(var r = 0; r < this.nrows; r++) {
	  if(!map[r] || n.addressType(map[r]) != 'dns') map[r] = col[r];
      }
      this.cdata[c] = map;
    }
  }
}

Table.prototype.collapseKeys = function(firstKeyIndex, nKeys, keyName, joinStr) {
    var lastKeyIndex = firstKeyIndex + nKeys - 1;
    var cnames = [];
    var ctypes = [];
    var kName = keyName || "key";
    var jStr = joinStr || "-";
    for(var cc in this.cnames) {
	if(cc == firstKeyIndex) {
	    cnames.push(kName);
	    ctypes.push("string");
	}
	else if(cc < firstKeyIndex || cc > lastKeyIndex) {
	    cnames.push(this.cnames[cc]);
	    ctypes.push(this.ctypes[cc]);
	}
    }
    var newTable = Table.create(cnames, ctypes);
    for(var rr = 0; rr < this.nrows; rr++) {
	var row = this.row(rr);
	var keyStr = null;
	if(row[firstKeyIndex] != null) {
	    var kk = [];
	    for(var cc = firstKeyIndex; cc <= lastKeyIndex; cc++) kk.push(row[cc]);
	    keyStr = kk.join(jStr);
	}
	var newRow = [];
	for(cc in this.cnames) {
	    if(cc == firstKeyIndex) newRow.push(keyStr);
	    else if(cc < firstKeyIndex || cc > lastKeyIndex) newRow.push(row[cc]);
	}
	newTable.addRow(newRow);
    }
    return newTable;
}

Table.prototype.printCSV = function(includeHeader,df,nf,lang,country) {
  if(includeHeader) {
    for(var c = 0; c < this.cdata.length; c++) {
      if(c > 0) print(",");
      var cname = this.cnames[c];
      if(cname) print(("" + cname).replace(/,/g,'\\u002c'));
    }
    println("");
  }
  for(var r = 0; r < this.nrows; r++) {
    for(var c = 0; c < this.ncols; c++) {
      if(c > 0) print(",");
      var val = this.cell(r,c);
      var pval = val;
      var type = typeof(val);
      if(typeof(val) == 'undefined' || val == null) pval = "";
      else if(type == 'object') { 
         if(val.constructor == Date) {
            if(df) pval = formatdate(val,df,lang,country);
            else pval = formatdate(val,"yyyyMMddHHmm");
         }
      }
      else if(type == 'number') {
         if(nf) pval = formatnumber(val,nf,lang,country);
      }
      print(("" + pval).replace(/,/g,'\\u002c'));
    }
    println("");
  }
} 

Table.prototype.printHTML = function(includeHeader,df,nf,lang,country) {
  println("<table>");
  if(includeHeader) {
    println("<thead>");
    print("<tr>");
    for(var c = 0; c < this.cdata.length; c++) {
      print("<th>");
      var cname = this.cnames[c];
      if(cname) print(cname); 
      print("</th>");
    }
    println("</tr>");
    println("</thead>");
  }
  println("<tbody>");
  for(var r = 0; r < this.nrows; r++) {
    print("<tr>");
    for(var c = 0; c < this.ncols; c++) {
      print("<td>");
      var val = this.cell(r,c);
      var pval = val;
      var type = typeof(val);
      if(typeof(val) == 'undefined' || val == null) pval = "";
      else if(type == 'object') {
         if(val.constructor == Date) {
           if(df) pval = formatdate(val,df,lang,country);
           else pval = val.toString();
         }
      }
      else if(type == 'number') {
         if(nf) pval = formatnumber(val,nf,lang,country);
         else pval = formatnumber(val,"#,###.##");
      }
      print(pval);

      print("</td>");
    }
    println("</tr>");
  }
  println("</tbody>");
  println("</table>");
}

Table.prototype.printFixed = function(includeHeader,widths,df,nf,lang,country) {

    var printfixed = function(string, width, center) {
        var w = Math.abs(width);
        if(string.length == w) print(string);
        else if(string.length > w) print(string.substr(0,w));
        else {
            var pad = w - string.length;
            if(center) {
                var lmargin = Math.floor(pad / 2);
                for(var i = 0; i < lmargin; i++) print(" ");
                print(string);
                var rmargin = pad - lmargin;
                for(var i = 0; i < rmargin; i++) print(" ");
            }
            else if(width >= 0) {
                print(string);
                for(var i = 0; i < pad; i++) print(" ");
            }
            else {
                for(var i = 0; i < (pad-1); i++) print(" ");
                print(string);
                print(" ");
            }
        }
    }

    var headstr = function(cname) {
        var phead = "";
        if(cname) phead = cname = cname.split(".").pop();
        return phead;
    }

    var valuestr = function(val) {
        var pval = val;
        var type = typeof(val);
        if(val == null) pval = "";
        else if(type == 'object') {
            if(val.constructor == Date) {
                if(df) pval = formatdate(val,df,lang,country);
                else pval = formatdate(val,"yyyyMMddHHmm");
            }
        }
        else if(type == 'number') {
            if(nf) pval = formatnumber(val,nf,lang,country);
        }
        return "" + pval;
    }

    var widtharr = null;
    if(widths) widtharr = widths;
    else {
        widtharr = new Array();

        // initialize, taking headers into account                                                                                                          
        for(var c = 0; c < this.cdata.length; c++) {
            widtharr[c] = 2;
            if(includeHeader) {
                var phead = headstr(this.cnames[c]);
                w = phead.length + 2;
                widtharr[c] = Math.max(widtharr[c], w );
            }
        }

        // now adapt to cell widths                                                                                                                         
        for(var r = 0; r < this.nrows; r++) {
            for(var c = 0; c < this.ncols; c++) {
                var val = this.cell(r, c);
                var pval = valuestr(val);
                var w = pval.length + 2;
                widtharr[c] = Math.max(widtharr[c],  w);


            }
        }
    }

    // now indicate numeric cols (by making widths negative)                                                                                                
    for(var c = 0; c < this.ncols; c++) {
        var type = this.ctypes[c];
        if(type == "integer" || type == "double") widtharr[c] = 0 - Math.abs(widtharr[c]);
    }

    // now actually print the table                                                                                                                         
    if(includeHeader) {
        for(var c = 0; c < this.ncols; c++) {
            var phead = headstr(this.cnames[c]);
            printfixed(phead,widtharr[c],true);
        }
        println("");
    }
    for(var r = 0; r < this.nrows; r++) {
        for(var c = 0; c < this.ncols; c++) {
            var val = this.cell(r,c);
            var pval = valuestr(val);
            printfixed(pval,widtharr[c],false);
        }
        println("");
    }
}

 
// Define classes for storing chart specifications
function XAxis() {
  this.colors = new Array();
}

function Series(name, values) {
  this.name   = name;
  this.values = values;
}

function YAxis() {
  this.series = new Array();
}
YAxis.prototype.addSeries = function(name, values) {
  var newSeries = new Series(name, values);
  this.series.push(newSeries);
}

function Chart(type) {
  this.type = type;
  this.xaxis = new XAxis();
  this.yaxis = new YAxis();
}

Chart.range = function(first,last) {
   var range = new Array();
   for(var i = first; i <= last; i++) range.push(i);
   return range;
}

Chart.multiSeries = function(type,title,table,xlabel,categoryColumn,ylabel,valueColumns) {
  var nseries = valueColumns.length;
                                                                                
  var chart = new Chart(type);
  chart.title = title;
  chart.start = table.start;
  chart.end = table.end;
  chart.group = table.group;

  chart.legend = nseries > 1;
                                                                                
  if(xlabel != null) chart.xaxis.label = xlabel;
  else chart.xaxis.label = table.cnames[categoryColumn];

  chart.xaxis.data = table.cdata[categoryColumn];
  chart.xaxis.type = table.ctypes[categoryColumn];
                                                                                
  if(ylabel != null) chart.yaxis.label = ylabel;
  else chart.yaxis.label = "value";
                                                                                
  for(var c = 0; c < nseries; c++) {
     chart.yaxis.addSeries(
       table.cnames[valueColumns[c]],
       table.cdata[valueColumns[c]]
     );
     if(table.pvttypes) {
       chart.yaxis.series[c].type = table.pvttypes[valueColumns[c]];
     }
  }
                                                                                
  return chart;
}

Chart.singleSeries = function(type,title, table, xlabel, categoryColumn, ylabel, valueColumn) {
  var valueColumns = new Array();
  valueColumns[0] = valueColumn;
  var yLabel = ylabel;
  if(yLabel == null) yLabel = table.cnames[valueColumn];
  return Chart.multiSeries(type,title, table, xlabel, categoryColumn, yLabel, valueColumns);
}

Chart.singleValue = function(type,title,ylabel,value,minValue,maxValue,marginalValue,criticalValue) {
  var chart = new Chart(type);
  chart.title = title;
  chart.legend = false;

  chart.xaxis.label = "";

  chart.xaxis.data = new Array();
  chart.xaxis.type = "string";

  chart.yaxis.label = ylabel;

  var valArr = new Array();
  valArr[0] = value;
  chart.xaxis.data[0] = "value";
  valArr[1] = minValue;
  chart.xaxis.data[1] = "minValue";
  valArr[2] = maxValue;
  chart.xaxis.data[2] = "maxValue";

  if(marginalValue && criticalValue) {
    valArr[3] = marginalValue;
    chart.xaxis.data[3] = "marginalValue";
    valArr[4] = criticalValue;
    chart.xaxis.data[4] = "criticalValue";
  }
  chart.yaxis.addSeries(
    "series",
    valArr
  );
  return chart;
}

// Report utility functions
 if(typeof Report != 'undefined') {
     Report.current = function() {
	 return new Report();
     }

     Report.prototype.tag = function(value,ttype,tval) {
	 return tag(value,ttype,tval);
     }
 }

// Widget utility functions
if(typeof Widget != 'undefined') {
    Widget.current = function() { return new Widget(); }
}

// WebApp utility functions
if(typeof WebApp != 'undefined') {
    WebApp.current = function() { return new WebApp(); }
}

// Network utility functions
Network.current = function() { return new Network(); }

Network.countMap = function(array) {
  var result = new Array();
  for(var i = 0; i < array.length; i++) {
    if(array[i]) {
      if(typeof(array[i]) == 'string') result[i] = 1;
      else if(array[i].length) result[i] = array[i].length;
      else result[i] = 1;
    }
    else result[i] = 0;
  }
  return result;
}

Network.uniqueMap = function(array,first) {
  var result = new Array(array.length);
  for(var i = 0; i < array.length; i++) {
    if(array[i]) {
      if(typeof(array[i]) == 'string') result[i] = array[i];
      else if(array[i].length) {
        if(first ? array[i].length >= 1 : array[i].length == 1) result[i] = array[i][0];
      }
      else result[i] = array[i];
    }
  }
  return result;
}

Network.prototype.allAgents = function() {
  if(!this._allagents) {
      var initialpath = this.path;
      this.path = null;
      this._allagents = this.agents();
      this.path = initialpath;
  }
  return this._allagents;
}

Network.prototype.allInterfaces = function() {
  if(!this._allinterfaces) {
      var initialpath = this.path;
      this.path = null;
      this._allinterfaces = this.interfaces();
      this.path = initialpath;
  }
  return this._allinterfaces;
}

Network.Inf = 100000;

Network.prototype.adjSet = function(useInterfaces) {
  if(!this._adjData) {
    var initialpath = this.path;
    var adj = {};
    var ifAdj = {};
    var ifs = this.allInterfaces();
    var neighbors = this.neighborMap(ifs);
    for(var i = 0; i < ifs.length; i++) {
      if(neighbors[i]) {
        this.path = ifs[i];
        var sagent = this.agentIP();
        this.path = neighbors[i];
        var dagent = this.agentIP();
        if(sagent && dagent) {
          adj[sagent + "," + dagent] = [ifs[i], neighbors[i]];
          ifAdj[ifs[i]] = neighbors[i];
        }
      }
    }
    this._adjData = [adj,ifAdj];
    this.path = initialpath;
  }
  return this._adjData[useInterfaces ? 1 : 0];
}

Network.prototype.adj = function(n1,n2,useInterfaces) {
  var res = this.adjSet()[n1 + "," + n2];
  return useInterfaces ? res : (res ? 1 : Network.Inf);
}

Network.prototype.shortestPaths = function(src) {
  if(!this._shortestPaths) this._shortestPaths = {};
  var paths = this._shortestPaths[src];
  if(!paths) {
     var agents = this.allAgents();
     var done = {};
     done[src] = true;
     var pathlengths = {};
     var predecessors = {};
     for(var i = 0; i < agents.length; i++) {
       var d = this.adj(src,agents[i]);
       pathlengths[agents[i]] =  d;
       if(d != Network.Inf) predecessors[agents[i]] = src;
     } 
     pathlengths[src] = 0;
     for(var i = 0; i < agents.length - 1; i++) {
       var closest = null;
       var closestDistance = Network.Inf;
       for (var j = 0; j < agents.length; j++) {
         if(!done[agents[j]] && pathlengths[agents[j]] < closestDistance) {
            closestDistance = pathlengths[agents[j]];
            closest = agents[j];
         }
       }
       done[closest] = true;
       for(var j = 0; j < agents.length; j++) {
         if(!done[agents[j]]) {
           var possiblyCloserDistance = pathlengths[closest] + this.adj(closest,agents[j]);
           if(possiblyCloserDistance < pathlengths[agents[j]]) {
             pathlengths[agents[j]] = possiblyCloserDistance;
             predecessors[agents[j]] = closest;
           }
         }
       }
     }
     paths = predecessors;
     this._shortestPaths[src] = paths;
  }
  return paths;
}

Network.prototype.shortestPath = function(destpath,useInterfaces) {
  if(!destpath) return null;
  var src = this.agentIP();
  if(!src) return null;
 
  var initialpath = this.path; 
  this.path = destpath;
  var dest = this.agentIP();
  this.path = initialpath;
  if(!dest || src == dest) {
     return null;
  }
  
  var predecessors = this.shortestPaths(src);
  var result;
  var p = predecessors[dest];
  if(p) {
    var pth = [dest];
    while(p != src) {
      pth.unshift(p);
      p = predecessors[p]
    }
    pth.unshift(src);
    if(useInterfaces) {
      var interfacesPath = [];
      for(var i = 1; i < pth.length; i++) {
        interfacesPath.push(this.adj(pth[i-1],pth[i],true));
      }
      result = interfacesPath;
    }
    else result = pth; 
  }
  else result = null;
  return result;
}

// Measurement submission
function Measurement(agent, ds_class, ds_index) {
  this.agent = agent;
  this.ds_class = ds_class || 0;
  this.ds_index = ds_index || 0x3FFFFFFF;
}

Measurement.prototype.counterSample = function(obj) {
  // splice in my meta-data
  var cs = {};
  cs.agent = this.agent;
  cs.ds_class = this.ds_class;
  cs.ds_index = this.ds_index;
  cs.counter_sample = obj;
  // then call into java
  measurement(JSON.stringify(cs));
}

Measurement.prototype.processor = function(tot_mem, free_mem, cpu5, cpu60, cpu300) {
  var cs = {};
  var pr = cs.processor = {};
  pr.total_memory = tot_mem;
  pr.free_memory = free_mem;
  pr.five_sec_cpu = cpu5;
  pr.one_min_cpu = cpu60;
  pr.five_min_cpu = cpu300;
  this.counterSample(cs);
}

Measurement.prototype.rtmetric = function(mname, mindex, mtype, mvalue) {
  var cs = {};
  var rtm = cs.rtmetric = {};
  rtm.mname = mname;
  rtm.mindex = mindex;
  rtm.mtype = mtype;
  rtm.mvalue = mvalue;
  this.counterSample(cs);
}
