-----------------------------------------------------------------------
----- Copyright (C) InMon Corp. 2014 ALL RIGHTS RESERVED       --------
----- Module to poll env data from hosts that support the      --------
----- ENTITY-SENSORS-MIB                                       --------
-----------------------------------------------------------------------
local json = require("dkjson")
local util = require("util")
local oid = require("oid")
local state = require("state")
local logger = require("logger")
local measurement = require("measurement")

local mod = "EntitySensors"

local stateDir = inx.pref("@StateDirectory")
local agentsDir = inx.pref("@AgentsDirectory")
local entity2 = "entity2.txt"

local e_hp = "1.3.6.1.4.1.11."
local e_arista = "1.3.6.1.4.1.30065."
local e_net_snmp = "1.3.6.1.4.1.8072."
local e_cisco = "1.3.6.1.4.1.9."

local entity_sensors = "1.3.6.1.2.1.99"

local es_sensor_type = entity_sensors .. ".1.1.1.1"
local es_sensor_scale = entity_sensors .. ".1.1.1.2"
local es_sensor_precision = entity_sensors .. ".1.1.1.3"
local es_sensor_value = entity_sensors .. ".1.1.1.4"
local es_sensor_status = entity_sensors .. ".1.1.1.5"

local entPhysClass_PSU = 6
local entPhysClass_Fan = 7

local function loadEntityTable(st, entityPath)
   -- build entPhysIndex -> { entPhysClass, entPhysContainedIn }
   -- and also count the PSUs and Fans
   st.psu_count = 0
   st.fan_count = 0
   io.input(entityPath)
   for line in io.lines() do
      local field = 0
      local physIdx = 0
      local containedIn = 0
      local physClass = 0
      local physDescr
      for tok in string.gmatch(line, '([^,]+)') do
	 if(field == 0) then physIdx = tonumber(tok) or 0 end
	 if(field == 1) then physDescr = tok end
	 if(field == 2) then containedIn = tonumber(tok) or 0 end
	 if(field == 3) then physClass = tonumber(tok) or 0 end
	 if(field >= 3) then break end
	 field = field + 1
      end
      if(physIdx > 0 and containedIn > 0 and physClass > 0) then
	 st.entityTable[physIdx] = {
	    entPhysDescr=physDescr,
	    entPhysIndex=physIdx,
	    entPhysClass=physClass,
	    entPhysContainedIn=containedIn
	 }
         if (physClass==entPhysClass_PSU) then st.psu_count = st.psu_count + 1
	 elseif (physClass==entPhysClass_Fan) then st.fan_count = st.fan_count + 1
	 end
      end
   end
end

local function readState(agent, st)
   local entityPath = stateDir .. "/" .. agentsDir .. "/" .. agent.ip .. "/" .. entity2
   local mtime = inx.modtime(entityPath)
   logger.dbg(mod, agent, "mtime of " .. entityPath .. " is " .. tostring(mtime))
   if not (mtime == st.entity_mtime) then
      st.entityTable = {}
      if not (mtime == 0) then
	 if(not pcall(loadEntityTable, st, entityPath)) then
	    logger.log(mod, agent, "exception in loadEntityTable")
	 end
      end
      st.entity_mtime = mtime
   end
end

local function parentEntityWithClass(st, entPhysIndex, searchClass)
   local pIdx = entPhysIndex
   while (st.entityTable[pIdx])  do
      local ent = st.entityTable[pIdx]
      if (ent.entPhysClass == searchClass) then	return pIdx end
      pIdx = ent.entPhysContainedIn
   end
   return nil
end

local function getEntityObj(st, oi, entPhysClass, key)
   local epIndex = oid.oid_index(oi)
   local objIndex = parentEntityWithClass(st, epIndex, entPhysClass)
   if (objIndex == nil) then
     return nil
   end
   if (not st[key]) then st[key] = {} end
   local lookup = st[key]
   if (not lookup[objIndex]) then lookup[objIndex] = { entPhysIndex=objIndex } end
   return lookup[objIndex]
end

local function printAgent(ip)
   local agent, st = state.getState(mod, ip)
   local etab = st.entityTable
   st.entityTable = "SUPPRESSED"
   logger.log(mod, agent, "printAgent:  " .. util.table_print(st))
   st.entityTable = etab
end

local function computeAgentPower(st)
   local psus = st.psus
   local id,psu
   for id,psu in pairs(psus) do
      if(psu.power) then
	 st.power = st.power + psu.power
      elseif (psu.voltage and psu.current) then
	 st.power = st.power + (psu.voltage * psu.current)
      end
   end
end

local function preWalkCB(ip)
   local agent,st = state.getState(mod, ip)
   if agent.abort then
      -- it failed last time so back off
      agent.poll_mS = agent.poll_mS * 5
   end
   readState(agent, st)
   st.abort = 0
   st.rows = 0
   st.power = 0
   st.temp_max = nil
   st.fan_max = nil
   st.humidity_max = nil

   st.fans = {}
   st.fan_failed = 0

   st.psus = {}
   st.psu_failed = 0
end

local function walkCB(ip, status, rowid, s_type, s_scale, s_precision, s_value, s_status)
   local agent,st = state.getState(mod, ip)
   logger.dbg(mod, agent, "walkCB: status=".. tostring(status))
   if status == nil then
      st.abort = 0
      st.rows = st.rows + 1
      -- see if this is for a PSU or Fan
      local psu = getEntityObj(st, rowid, entPhysClass_PSU, "psus")
      local fan = getEntityObj(st, rowid, entPhysClass_Fan, "fans")
      -- test the sensor_status first
      local sensor_status = tonumber(s_status) or 0
      if(sensor_status == 3) then
	 if (psu and psu.failed == nil) then
	    psu.failed = 1
	    st.psu_failed = st.psu_failed + 1
	 else
	    if(fan and fan.failed == nil) then
	       fan.failed = 1
	       st.fan_failed = st.fan_failed + 1
	    end
	 end
      elseif(sensor_status == 1) then
	 -- sensor_status is OK, so accept the reading (ignore sensor_status == 2)

	 if (psu and oid.objIdMatch(agent.objId, { e_arista })) then
	    -- arista special case: the PSUs have both input and output sensors for
	    -- voltage and current, so here we ignore output sensors to ensure we
	    -- get a clean signal for input power.  Otherwise we would be multiplying
	    -- max(in_volts, out_volts) * max(in_current, out_current). Which might
	    -- be wierd.
	    local ent = st.entityTable[psu.entPhysIndex]
	    if(ent and ent.entPhysDescr and string.find(ent.entPhysDescr, "output")) then
	       psu = nil
	    end
	 end

	 local sensor_type = tonumber(s_type)
	 local sensor_scale = 10^((9 - tonumber(s_scale)) * 3) -- e.g. 8=>milli, 7=>micro
	 local sensor_precision = 10^(tonumber(s_precision)) -- e.g. 0=>1, 1=>0.1 increments
	 local scaled_val = sensor_scale / sensor_precision * (tonumber(s_value) or 0)
	 -- now interpret sensor_type
	 if(sensor_type == 3 or sensor_type == 4) then -- volts (AC or DC)
	    if (psu) then measurement.accumulate(psu, "voltage", scaled_val, "max") end
	 elseif(sensor_type == 5) then -- Amperes
	    if (psu) then measurement.accumulate(psu, "current", scaled_val, "max") end
	 elseif(sensor_type == 6) then -- Watts
	    if (psu) then measurement.accumulate(psu, "power", scaled_val, "max") end
	 elseif(sensor_type == 8) then -- Celcius
	    measurement.accumulate(st, "temp_max", scaled_val, "max")
	 elseif(sensor_type == 9) then -- Relative Humidity%
	    measurement.accumulate(st, "humidity_max", scaled_val, "max")
	 elseif(sensor_type == 10) then -- fan rpm
	    if (fan) then measurement.accumulate(st, "fan_max", scaled_val, "max") end
	 end
      end
   elseif (status == "END") then
      if not pcall(computeAgentPower, st) then
	 logger.log(mod, agent, "computeAgentPower failed")
      end

      if(agent.debug) then printAgent(ip) end
      
      measurement.env(ip,
		      st.power,
		      st.temp_max,
		      st.humidity_max,
		      st.fan_max,
		      st.fan_count,
		      st.fan_failed,
		      st.psu_count,
		      st.psu_failed)
   elseif (status == "ABORT") then
      st.abort = st.abort + 1
   end
   return true -- keep going
end

return {
   mod = mod,
   objIds = { e_hp, e_arista, e_net_snmp, e_cisco },
   oids = { es_sensor_type, es_sensor_scale, es_sensor_precision, es_sensor_value, es_sensor_status },
   preWalkCB = preWalkCB,
   walkCB = walkCB
}
