-----------------------------------------------------------------------
----- Copyright (C) InMon Corp. 2016 ALL RIGHTS RESERVED           ----
----- Module to poll Optical (DOM) stats from Brocade(TM) devices  ----
-----------------------------------------------------------------------

local measurement = require("measurement")
local logger = require("logger")
local state = require("state")
local util = require("util")
local oid = require("oid")

local mod = "Brocade_Optical"

local e_brocade = "1.3.6.1.4.1.1991."
local e_brocade_MLX = e_brocade .. "1.3.44"
local e_brocade_MLXe = e_brocade .. "1.3.55"

-- for multi-laser QSFPs and beyond there is a new table with an entry per lane
-- local e_brocade_snIfOpticalLaneMonitoringTable = e_brocade .. "1.1.3.3.10."

-- but for now we assume one laser and use this table
local e_brocade_snIfOpticalMonitoringTable = e_brocade .. "1.1.3.3.6."
local e_sfp_temp = e_brocade_snIfOpticalMonitoringTable .. "1.1"
local e_sfp_tx_pwr = e_brocade_snIfOpticalMonitoringTable .. "1.2"
local e_sfp_rx_pwr = e_brocade_snIfOpticalMonitoringTable .. "1.3"
local e_sfp_tx_bias = e_brocade_snIfOpticalMonitoringTable .. "1.4"

local function firstTokenAsNumber(agent, field, str)
      if str == nil then return 0 end
      local num = str:match("([+%-0-9.eE]+)") -- float
      logger.dbg(mod, agent, "firstTokenAsNumber " .. field .. " extract " .. str .. " -> " .. tostring(num));
      return tonumber(num) or 0
end

local function dBm_to_mW(dBm)
      return math.pow(10, (dBm / 10.0))
end

local function preWalkCB(ip)
   local agent,st = state.getState(mod,ip)
   st.interfaces = {}
end

local function getInterface(st, ifindex)
   if not st.interfaces[ifindex] then
      st.interfaces[ifindex] = { lanes = {}, num_lanes = 0 }
   end
   return st.interfaces[ifindex]
end

local function walkCB(ip, status, rowid, temp_s, tx_pwr_s, rx_pwr_s, tx_bias_s)
   local agent,st = state.getState(mod,ip)

   if (status == nil) then
      -- rowid may be <ifindex> or <ifindex>.<laneid>
      local array = oid.oid_array(rowid)
      local ifindex = array[1]
      local laneId = array[2] or 1
      local intf = getInterface(st, ifindex)
      intf.mV = 0 -- supply voltage
      intf.nM = 0 -- wavelength
      intf.degC = firstTokenAsNumber(agent, "temp", temp_s)
      local tx_dBm = firstTokenAsNumber(agent, "tx_pwr", tx_pwr_s)
      local rx_dBm = firstTokenAsNumber(agent, "rx_pwr", rx_pwr_s)
      local tx_mA = firstTokenAsNumber(agent, "tx_bias_curr", tx_bias_s)
      logger.dbg(mod, agent, "rowid=" .. rowid .. " rx_dBm=" .. tostring(rx_dBm) .. " r_mA=" .. tostring(rx_mA))
      if intf.degC ~= 0 or tx_dBm ~= 0 or rx_dBm ~= 0 or tx_mA ~= 0 then
         local lane = measurement.sfp_lane(laneId,
		                            1000.0 * dBm_to_mW(tx_dBm),
		     		            1000.0 * dBm_to_mW(rx_dBm),
					    1000.0 * tx_mA,
					    intf.nM)
         intf.lanes[laneId] = lane
	 intf.num_lanes = intf.num_lanes + 1
         logger.dbg(mod, agent, "rowid=" .. rowid .. " lane=" .. util.table_print(lane, 2) .. " tx_dBm=" .. tostring(tx_dBm))
      end
   elseif (status == "END") then
      -- push out the results for all interfaces
      for idx,intf in pairs(st.interfaces) do
         if (intf.num_lanes > 0) then
            measurement.sfp(ip, idx, (1000 * intf.degC), intf.mV, intf.num_lanes, intf.lanes)
         end
      end
      st.interfaces = {} -- reset for next walk
   end
   return true -- keep going
end

return {
   mod = mod,
   objIds = { e_brocade_MLX, e_brocade_MLXe },
   oids = { e_sfp_temp, e_sfp_tx_pwr, e_sfp_rx_pwr, e_sfp_tx_bias },
   preWalkCB = preWalkCB,
   walkCB = walkCB
}
