-----------------------------------------------------------------------
----- Copyright (C) InMon Corp. 2014 ALL RIGHTS RESERVED       --------
----- Module to collect extra metrics from agents              --------
-----------------------------------------------------------------------
local util = require("util")

-- debug - set both of these to ip addresses
local debugIP1 = inx.pref("@LOCDebugAgent1")
local debugIP2 = inx.pref("@LOCDebugAgent2")
local debugLsid = ""
-- or this to addr>ifindex
local debugPid = inx.pref("@LOCDebugPort") or ""
   
-- files
local stateDir = inx.pref("@StateDirectory")
local linksDirName = inx.pref("@TPLinks")
local linkedPortsDirName = inx.pref("@TPLinkedPorts")
local linksDir = stateDir .. "/" .. linksDirName
local linkedPortsDir = stateDir .. "/" .. linkedPortsDirName
local topologyFile = stateDir .. "/" .. inx.pref("@Topology")
local topologyChangeFile = stateDir .. "/" .. inx.pref("@TopologyChange")
local agentsFile = stateDir .. "/" .. inx.pref("@AgentsFile")
local interfacesFile = stateDir .. "/" .. inx.pref("@InterfacesFile")
local candidatesFile = stateDir .. "/" .. inx.pref("@CLinks")

-- controls
local linkTypeFilter = util.prefBoolean(inx.pref("@LinkTypeFilter"))
local linkSpeedFilter = util.prefBoolean(inx.pref("@LinkSpeedFilter"))
local linkStatusFilter = util.prefBoolean(inx.pref("@LinkStatusFilter"))
local raiseEvent = util.prefBoolean(inx.pref("@event.interface.link"))
local timeout_hours = inx.pref("@TopologyLinkTimeoutHours")
timeout_hours = tonumber(timeout_hours) or 24
local timeout = timeout_hours * 3600

-- module vars
local agents={} -- key by ip
local ports={}  -- key by ip,ifIndex
local links={}  -- array
local linkSets={} -- key by ip-pair
local lagPorts={}
local lagLinks={}
local links_added = 0
local links_removed = 0

------------------------------------------------
-------- functions to assemble datamodel -------
------------------------------------------------

local function getAgent(atab, ip)
   -- find or create the agent
   if(not atab[ip]) then atab[ip] = { ip=ip, ports={}, v2p={} }  end
   return atab[ip]
end

local function portID(ip, ifIndex)
   return ip .. ">" .. ifIndex
end

local function getPort(ptab, agent, ifIndex)
   -- find or create the port
   local pid = portID(agent.ip, ifIndex)
   if(not ptab[pid]) then
      ptab[pid] = { pid=pid, agent=agent, ifIndex=ifIndex, flags=0 }
      if(pid==debugPid) then ptab[pid].log = true end
   end
   return ptab[pid]
end

local function linkID(port1, port2)
   -- sort the keys
   if(port1.pid > port2.pid) then port1,port2 = port2,port1 end
   return port1, port2, port1.pid .. "-" .. port2.pid
end

local function getLink(ltab, port1, port2)
   -- find or create the link
   local p1, p2, lid = linkID(port1, port2)
   if(not ltab[lid]) then ltab[lid] = { lid=lid, port1=p1, port2=p2, score=0, flags=0, algs={}}  end
   return ltab[lid]
end

local function linkSetID(ip1, ip2)
   -- sort keys
   if(ip1 > ip2) then ip1,ip2 = ip2,ip1 end
   return ip1,ip2,ip1 .. "-" .. ip2
end

local function getLinkSet(lstab, agent1, agent2)
   -- find or create the linkSet
   local ip1,ip2,lsid = linkSetID(agent1.ip, agent2.ip)
   if(ip1 == agent2.ip) then agent1,agent2 = agent2,agent1 end
   if(not lstab[lsid]) then
      lstab[lsid] = { lsid=lsid, agent1=agent1, agent2=agent2, lports={}, n_lports=0, links={} }
      if(lsid==debugLsid) then lstab[lsid].log = true end
   end
   return lstab[lsid]
end

local function addPartial(lset, port)
   if(not lset.lports[port.pid]) then
      lset.lports[port.pid] = port
      lset.n_lports = lset.n_lports + 1
   end
end

local function removePartial(lset, port)
   if(lset.lports[port.pid]) then
      lset.lports[port.pid] = nil
      lset.n_lports = lset.n_lports - 1
   end
end

local function lsetLog(lset, str)
   if(lset.log) then
      -- break into lines to allow large str
      for line in string.gmatch(str, "[^\n]+") do
	 inx.log(0, "linkSet(".. lset.lsid .. "): " .. line)
      end
   end
end

local function prtLog(prt, str)
   if(prt.log) then
      -- break into lines to allow large str
      for line in string.gmatch(str, "[^\n]+") do
	 inx.log(0, "port(".. prt.pid .. "): " .. line)
      end
   end
end

local function accumulate(tab, field, val)
   if(not tab[field]) then tab[field] = 0 end
   tab[field] = tab[field] + val
end

------------------------------------------------
------------- addLink --------------------------
------------------------------------------------

local function addLink(ip1,ifIndex1,ip2,ifIndex2,lastSeen,incumbent,score,flags,alg)
   lastSeen = tonumber(lastSeen) or 0
   local ag1 = agents[ip1]
   local ag2 = agents[ip2]
   if(ag1 and ag2 and ((inx.now() - lastSeen) < timeout)) then
      ifIndex1 = tonumber(ifIndex1) or 0
      ifIndex2 = tonumber(ifIndex2) or 0
      -- map back from virtual to physical early in case some
      -- algorithms are dealing in virtual ports and others are
      -- dealing in physical ones
      if(ag1.v2p[ifIndex1]) then ifIndex1 = ag1.v2p[ifIndex1] end
      if(ag2.v2p[ifIndex2]) then ifIndex2 = ag2.v2p[ifIndex2] end
      local p1 = nil
      local p2 = nil
      if(ifIndex1 ~= 0) then p1 = getPort(ag1.ports, ag1, ifIndex1) end
      if(ifIndex2 ~= 0) then p2 = getPort(ag2.ports, ag2, ifIndex2) end
      local lset = getLinkSet(linkSets, ag1, ag2)
      if(p1 and p2) then
	 -- fully-specified links
	 local lnk = getLink(lset.links, p1, p2)
	 lnk.flags = inx.bitOP(lnk.flags, "|", flags)
	 if(incumbent) then
	    lnk.incumbent = true
	 end
	 lnk.score = lnk.score + score
	 accumulate(lnk.algs,alg,score)
	 lnk.linkSet = lset
	 prtLog(p1, "link candidate: " .. lnk.lid .. " alg=" .. alg .. " score=" .. score)
	 prtLog(p2, "link candidate: " .. lnk.lid .. " alg=" .. alg .. " score=" .. score)
      else
	 -- partial links
	 if(p1) then
	    addPartial(lset, p1)
	    prtLog(p1, "partially specified link " .. lset.lsid)
	 end
	 if(p2) then
	    addPartial(lset, p2)
	    prtLog(p2, "partially specified link " .. lset.lsid)
	 end
      end
   end
end

local function delete_v2p()
   for ip,ag in pairs(agents) do
      ag.v2p = nil
   end
end

------------------------------------------------
------------- lines from agents file  ----------
------------------------------------------------

local function processAgentsLine(args, agentIP,contactIP,community,writeCommunity,enterprise,site,zone,subnetName,agentName,sysName,dnsName,icmpOK,icmpTested,snmpOK,snmpTested,sourceTypeFlows,lastTalkedFlows,sourceTypeCounters,lastTalkedCounters,poll,bridgePortOK,bridgePortTested,bridgeOK,bridgeTested,bgpPeer,sflowOK,sflowTested,xrmonOK,xrmonTested,ifTableOK,ifTableTested,sflowMibOK,sFlowMibTested,sFlowMibIncomplete,xrmonMibOK,xrmonMibTested,xrmonMibIncomplete,agentOrRange,mibForce,clash,configIP,csiOK,csiTested,discoverySource,sysObjectID,dsFlags,samplingDirection,dsFlagsCtr,flowDiscoveryIndex,inferPorts,retest,downFlag,lastTalkedICMP,lastTalkedSNMP,hostSection,trunkFlags)
   if(agentIP) then
      local ag = getAgent(agents, agentIP)
      ag.sysObjectID = sysObjectID
      ag.sysName = sysName
   end
end

------------------------------------------------
---- lines from linkedPorts directory ----------
------------------------------------------------

local function processLinkedPortsLine(args, agentIP,ifIndex,lastSeen,mod,flags,flagsStr)
   -- load ports for existing agents
   local ag = agents[agentIP]
   if(ag) then
      local idx = tonumber(ifIndex) or 0
      -- map back to physical
      if(ag.v2p[idx]) then idx = ag.v2p[idx] end
      local prt = getPort(ag.ports, ag, idx)
      prt.flags = inx.bitOP(prt.flags, "|", tonumber(flags) or 0)
   end
end

local function readLinkedPorts(file, args)
   for line in io.lines(linkedPortsDir .. "/" .. file) do
      inx.tokensDo(line, ",", true, processLinkedPortsLine, args)
   end
end

------------------------------------------------
--------- lines from interfaces file  ----------
------------------------------------------------

local function operStatusUp(ifStatus)
   ifStatus = tonumber(ifStatus) or 0
   return inx.bitOP(ifStatus, ">", 1)
end

local function processInterfacesLine1(args, agentIP,ifIndex,mac,ipAddr,subnet,lastSeen,bridgePort,ifName,ifDescr,ifType,ifSpeed,physIfIndex,vlan,vlanLastSeen,vlanFlags,lagIfIndex,ifAlias,sa_ifSpeed,sa_ifName,ifStatus)
   if(not agentIP) then return end
   -- build v2p lookups
   local ag = agents[agentIP]
   if(ag) then
      local idx = tonumber(ifIndex) or 0
      local phy = tonumber(physIfIndex) or 0
      if(phy > 0 and phy ~= idx) then
	 ag.v2p[idx] = phy
      end
   end
end

local function processInterfacesLine2(args, agentIP,ifIndex,mac,ipAddr,subnet,lastSeen,bridgePort,ifName,ifDescr,ifType,ifSpeed,physIfIndex,vlan,vlanLastSeen,vlanFlags,lagIfIndex,ifAlias,sa_ifSpeed,sa_ifName,ifStatus)
   if(not agentIP) then return end
   -- add details to existing ports.  Should be linked (or suspected) physical ports only.
   local ag = agents[agentIP]
   if(ag) then
      local idx = tonumber(ifIndex) or 0
      local pid = portID(agentIP,idx)
      local pt = ag.ports[pid]
      if(pt) then
	 pt.ifName = ifName
	 pt.ifDescr = ifDescr
	 pt.ifAlias = ifAlias
	 pt.ifType = ifType
	 pt.ifSpeed = tonumber(ifSpeed) or 0
	 pt.lagIfIndex = tonumber(lagIfIndex) or 0
	 pt.up = operStatusUp(ifStatus)
	 if(pt.log) then prtLog(pt, "details added: " .. util.table_print(pt, 1)) end
      end
   end
end

------------------------------------------------
----------- lines from topology file  ----------
------------------------------------------------

local function processTopologyLine(args, agent1,ifIndex1,agent2,ifIndex2,firstSeen,lastSeen,score,lag)
   local aggID = tonumber(lag) or 0
   if(agent1 and agent2 and aggID == 0) then 
      addLink(agent1,ifIndex1,agent2,ifIndex2,lastSeen,true,1,0,"inc")
   end
end

------------------------------------------------
---------- lines from links directory ----------
------------------------------------------------

local function processLinksLine(args,test,agent1,ifIndex1,agent2,ifIndex2,firstSeen,lastSeen,flags,alg,w8ag1,w8if1,w8ag2,w8if2,flagsStr)
   if (agent1 and agent2 and ifIndex1 and ifIndex2) then
      local score = w8ag1 + w8if1 + w8ag2 + w8if2
      addLink(agent1,ifIndex1,agent2,ifIndex2,lastSeen,false,score,flags,alg)
   end
end

local function readLinks(file, args)
   for line in io.lines(linksDir .. "/" .. file) do
      inx.tokensDo(line, ",", true, processLinksLine, args)
   end
end

------------------------------------------------
----------------- portAffinity   ---------------
------------------------------------------------

local function portAffinity(port1, port2)
   local aff = 0
   if(port1.ifSpeed == port2.ifSpeed) then aff = aff + 100 end
   if(port1.up == port2.up) then aff = aff + 10 end
   if(port1.ifType == port2.ifType) then aff = aff + 1 end
   -- hard overrides for link*Filter settings
   if(linkStatusFilter and ((port1.up) ~= (port2.up))) then aff = -1 end
   if(linkSpeedFilter and ((port1.ifSpeed) ~= (port2.ifSpeed))) then aff = -1 end
   if(linkTypeFilter and ((port1.ifType) ~= (port2.ifType))) then aff = -1 end
   return aff
end

------------------------------------------------
--------------- resolveLinkSets  ---------------
------------------------------------------------

local function resolveLinkSet(lset)
   -- another algorithm may have offered a complete link resolution for
   -- one of the partial ones.  If so, just remove the partial.
   if(lset.log) then lsetLog(lset, util.table_print(lset, 3)) end
   for lid,lnk in pairs(lset.links) do
      removePartial(lset, lnk.port1)
      removePartial(lset, lnk.port2)
   end

   -- try to pair up partial ports to form new candidate links
   if(lset.n_lports > 1) then
      -- separate into "left" and "right" arrays by agent ip
      lsetLog(lset, "merge partials: " .. lset.n_lports)
      local ag_left = nil
      local left = {}
      local right = {}
      local n_left, n_right = 0,0
      for pid,prt in pairs(lset.lports) do
	 if(not ag_left) then ag_left = prt.agent end
	 local side
	 if(prt.agent.ip == ag_left.ip) then
	    side = left
	    n_left = n_left + 1
	 else
	    side = right
	    n_right = n_right + 1
	 end
	 table.insert(side, prt)
      end
      -- iterate, promoting closest pair to candidate link assertion
      while (table.getn(left) > 0 and
	     table.getn(right) > 0) do
	 local closest = nil
	 local affinity = -1
	 for ii,prt1 in ipairs(left) do
	    for jj,prt2 in ipairs(right) do
	       local aff = portAffinity(prt1, prt2)
	       if(aff > affinity) then
		  closest = { prt1, prt2, ii, jj }
		  affinity = aff
	       end
	    end
	 end
	 local pmg1 = closest[1]
	 local pmg2 = closest[2]
	 local merged = getLink(lset.links, pmg1, pmg2)
	 lsetLog(lset, "resolveLinkSet: merged link: " .. merged.lid)
	 prtLog(pmg1, "resolveLinkSet: merged link: " .. merged.lid)
	 prtLog(pmg2, "resolveLinkSet: merged link: " .. merged.lid)
	 merged.score = merged.score + 1
	 merged.flags = inx.bitOP(merged.flags, "|", pmg1.flags)
	 merged.flags = inx.bitOP(merged.flags, "|", pmg2.flags)
	 merged.linkSet = lset
	 table.remove(left, closest[3])
	 table.remove(right, closest[4])
      end
      -- may have an odd one out left over
      if(n_left ~= n_right) then
	 lsetLog(lset, "odd partial (left=" .. n_left .. ", right=" .. n_right .. ")")
      end
      -- also infer LAG relationship
      if(n_left > 1 and n_right > 1) then
	 local comp1 = nil
	 local comp2 = nil
	 for lid,lnk in pairs(lset.links) do
	    local lag1 = lnk.port1.lagIfIndex or 0
	    local lag2 = lnk.port2.lagIfIndex or 0
	    if(lag1 > 0) then comp1 = lnk.port1 end
	    if(lag2 > 0) then comp2 = lnk.port2 end
	 end
	 if(comp1 and comp2) then
	    -- create separate collection of ports and links for LAGs
	    local aggPort1 = getPort(lagPorts, comp1.agent, comp1.lagIfIndex)
	    local aggPort2 = getPort(lagPorts, comp2.agent, comp2.lagIfIndex)
	    local lag = getLink(lagLinks, aggPort1, aggPort2)
	    -- should we set score, flags and algs?
	    lsetLog(lset, "added lag link: " .. lag.lid)
	    prtLog(aggPort1, "added lag link: " .. lag.lid)
	    prtLog(aggPort2, "added lag link: " .. lag.lid)
	 end
      end
   end
end

local function resolveLinkSets()
   for lsid,lset in pairs(linkSets) do
      resolveLinkSet(lset)
   end
end

------------------------------------------------
----------------- output utils  ----------------
------------------------------------------------

local function flagsStr(flags)
   local str = ""
   if(inx.bitOP(flags, "&", 0x00001) > 0) then str = str .. " USR" end
   if(inx.bitOP(flags, "&", 0x00002) > 0) then str = str .. " AR2" end
   if(inx.bitOP(flags, "&", 0x00004) > 0) then str = str .. " AR1" end
   if(inx.bitOP(flags, "&", 0x00008) > 0) then str = str .. " LOC" end
   if(inx.bitOP(flags, "&", 0x00010) > 0) then str = str .. " SUB" end
   if(inx.bitOP(flags, "&", 0x00020) > 0) then str = str .. " CTR" end
   if(inx.bitOP(flags, "&", 0x00040) > 0) then str = str .. " XDP" end
   if(inx.bitOP(flags, "&", 0x00080) > 0) then str = str .. " STP" end
   if(inx.bitOP(flags, "&", 0x00100) > 0) then str = str .. " RMC" end
   if(inx.bitOP(flags, "&", 0x00200) > 0) then str = str .. " STC" end
   if(inx.bitOP(flags, "&", 0x00400) > 0) then str = str .. " XDC" end
   if(inx.bitOP(flags, "&", 0x00800) > 0) then str = str .. " RFT" end
   if(inx.bitOP(flags, "&", 0x01000) > 0) then str = str .. " MGT" end
   if(inx.bitOP(flags, "&", 0x02000) > 0) then str = str .. " PMC" end
   if(inx.bitOP(flags, "&", 0x04000) > 0) then str = str .. " DCB" end
   if(inx.bitOP(flags, "&", 0x08000) > 0) then str = str .. " LAG" end
   if(inx.bitOP(flags, "&", 0x10000) > 0) then str = str .. " INC" end
   return str
end

local function algStr(algs)
   local str = ""
   if(algs) then
      for alg,score in pairs(algs) do
	 str = str .. " " .. alg .. "=" .. score
      end
   end
   return str
end

local function writeLink(ff, p1, p2, score, lag, flags, algs)
   local ip1,ip2,if1,if2,code = "","",0,0,""
   local fl = tonumber(flags) or 0
   if(p1) then
      ip1 = p1.agent.ip
      if1 = p1.ifIndex
   end
   if(p2) then
      ip2 = p2.agent.ip
      if2 = p2.ifIndex
   end
   ff:write(ip1, ",", if1, ",", ip2, ",", if2, ",", 0, ",", inx.now(), ",", (score or 0), ",", (lag or ""), ",", fl, ",", flagsStr(fl), "," , algStr(algs), "\n")
end

------------------------------------------------
---------- writeCandidateLinks  ----------------
------------------------------------------------

local function writeCandidateLinks()
   local buildFile = candidatesFile .. ".bld"
   local ff = io.open(buildFile, "w")
   for lsid,lset in pairs(linkSets) do
      -- fully specified
      for lid,lnk in pairs(lset.links) do
	 writeLink(ff, lnk.port1, lnk.port2, lnk.score, "", lnk.flags, lnk.algs)
      end
      -- and partial
      for pid,prt in pairs(lset.lports) do
	 local rem_ag = lset.agent1
	 if(prt.agent.ip == lset.agent1.ip) then rem_ag = lset.agent2 end
	 local partial = getPort({}, rem_ag, 0)
	 writeLink(ff, prt, partial, 0, "", prt.flags)
      end
   end
   ff:close()
   os.rename(buildFile, candidatesFile)
end

------------------------------------------------
------------------ assignLinks  ----------------
------------------------------------------------

local function compareLinkScore(lnk1, lnk2)
   return (lnk1.score > lnk2.score)
end

local function assignLinks()
   local linkArray = {}
   for lsid,lset in pairs(linkSets) do
      for lid,lnk in pairs(lset.links) do
	 table.insert(linkArray, lnk)
      end
   end
   table.sort(linkArray, compareLinkScore)
   for ii,lnk in ipairs(linkArray) do
      local preemption = lnk.port1.link or lnk.port2.link
      local affinity = portAffinity(lnk.port1, lnk.port2)
      if(preemption or (lnk.score < 0) or (affinity < 0)) then
	 -- remove link
	 if(lnk.linkSet and lnk.linkSet.log) then
	    local prefix = "assignLinks: link = " .. lnk.lid 
	    if (preemption) then
	       inx.log(0, prefix .. " score=" .. lnk.score .. " preempted by " .. preemption.lid .. " score=" .. preemption.score)
	    elseif(lnk.score < 0) then
	       inx.log(0, prefix .. " negative score: " .. lnk.score)
	    elseif(affinity < 0) then
	       inx.log(0, prefix .. " negative port affinity: " .. affinity)
	    end
	 end
	 if(lnk.incumbent) then links_removed = links_removed + 1 end
      else
	 -- acquire ports
	 lnk.port1.link = lnk
	 lnk.port2.link = lnk
	 -- and add to the final list
	 table.insert(links, lnk)
      end
   end
end

------------------------------------------------
---------------- writeTopology  ----------------
------------------------------------------------

local function writeTopology()
   inx.log(1, "writeTopology links=" .. tostring(table.getn(links)))
   local buildFile = topologyFile .. ".bld"
   local ff = io.open(buildFile, "w")
   for ii,lnk in ipairs(links) do
      if(not lnk.incumbent) then
	 links_added = links_added + 1
	 if raiseEvent then
	    inx.raiseEvent("inform","interface","link",lnk.lid,"", lnk.port1.agent.ip, 0, lnk.port1.ifIndex, nil)
         end
      end
      writeLink(ff, lnk.port1, lnk.port2, lnk.score, "", lnk.flags, lnk.algs)
   end
   -- indicate other "linked ports"
   for ip,ag in pairs(agents) do
      for pid,prt in pairs(ag.ports) do
	 if(not prt.link) then
	    writeLink(ff, prt, nil, 0, "", prt.flags, nil)
	 end
      end
   end
   -- and indicate "lag associations"
   for lid,lnk in pairs(lagLinks) do
      writeLink(ff, lnk.port1, lnk.port2, lnk.score, "L", lnk.flags, nil)
   end
   ff:close()
   os.rename(buildFile, topologyFile)
end

------------------------------------------------
----------- writeTopologyChange  ---------------
------------------------------------------------

local function writeTopologyChange()
   local ff = io.open(topologyChangeFile, "w")
   ff:write("+",links_added,",-",links_removed,"\n")
   ff:close()
end

------------------------------------------------
------------- resolveTopology ------------------
------------------------------------------------


local function resolveTopology(args)
   local line
   local args = {}

   -- agents
   for line in io.lines(agentsFile) do
      inx.tokensDo(line, ",", true, processAgentsLine, args)
   end

   -- build virtual to physical lookups
   if(inx.modtime(interfacesFile) > 0) then
      for line in io.lines(interfacesFile) do
	 inx.tokensDo(line, ",", true, processInterfacesLine1, args)
      end
   end

   -- linked ports
   inx.filesDo(linkedPortsDir, readLinkedPorts, args )

   -- topology
   if(inx.modtime(topologyFile) > 0) then
      for line in io.lines(topologyFile) do
	 inx.tokensDo(line, ",", true, processTopologyLine, args)
      end
   end

   -- candidate links
   inx.filesDo(linksDir, readLinks, args )

   -- read in extra interface info
   if(inx.modtime(interfacesFile) > 0) then
      for line in io.lines(interfacesFile) do
	 inx.tokensDo(line, ",", true, processInterfacesLine2, args)
      end
   end

   -- drop v2p just so util.table_print is sane
   delete_v2p()

   -- merge partial links within linkSets
   resolveLinkSets()

   -- write candidates (for other modules to comment on)
   writeCandidateLinks()

   -- sort links by score and assign them to ports
   assignLinks()
end

------------------------------------------------
------------------------------------------------
------------------------------------------------
if(debugIP1 and debugIP2) then
   debugIP1,debugIP2,debugLsid = linkSetID(debugIP1, debugIP2)
end

resolveTopology()
writeTopology()
writeTopologyChange()

-- signal that we are done
local endEvent = inx.sourceEvent("end", "end event from Lua");
inx.sendEvent(endEvent, "all done");

local topology={
   agents=agents,
   ports=ports,
   links=links,
   linkSets=linkSets,
   lagsPorts=lagPorts,
   lagLinks=lagLinks
}

return topology
